
import numpy as np
from arnica.utils.curve_feat_extract import mask_boundary_layer, reptile, amax_keepshape, amin_keepshape


def test_mask_boundary_layer(allclose):
    """Damo of a simplified set of graphs with showy"""
    source=  np.array([
        [0,1,0,0,0],
        [0,1,1,1,1],
        [0,3,2,3,4],
        [0,1,-2,3,4],
        [0,0,0,0,0],
        [0,1,2,3,4],
    ])

    target = np.array(
        [[1, 1, 0, 0, 0],
        [1, 0, 0, 0, 0],
        [1, 1, 0, 0, 0],
        [1, 1, 1, 0, 0],
        [1, 0, 0, 0, 0],
        [1, 0, 0, 0, 0]]
    )

    assert allclose(target,mask_boundary_layer(source, eps=0.1, axis=1))

def test_reptile(allclose):
    """Damo of a simplified set of graphs with showy"""
    source=  np.array(
        [[0, 1, 0, 0, 0],
        [0, 1, 1, 1, 1],
        [0, 1, 2, 3, 4]])

    target = np.array(
        [[0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1],
        [3, 3, 3, 3, 3]]
    )

    assert allclose(target,reptile(source, pos=3, axis=1))

def test_amax_keepshape(allclose):
    """Damo of a simplified set of graphs with showy"""
    source=  np.array(
        [[0, 1, 0, 0, 0],
        [0, 1, 1, 1, 1],
        [0, 1, 2, 3, 3]])

    target = np.array(
        [[1, 1, 1, 1, 1],
          [1, 1, 1, 1, 1],
          [3, 3, 3, 3, 3]]
    )

    assert allclose(target,amax_keepshape(source,  axis=1))

def test_amin_keepshape(allclose):
    """Damo of a simplified set of graphs with showy"""
    source=  np.array(
        [[0, 1, 0, 0, 0],
        [0, 1, -1, 1, 1],
        [1, 1, 2, 3, 4]])

    target = np.array(
        [[0, 0, 0, 0, 0],
        [-1, -1, -1, -1, -1],
        [1, 1, 1, 1, 1]]
    )

    assert allclose(target,amin_keepshape(source,  axis=1))


""" module to test sample_curve_from_cloud.py of Arnica"""

import numpy as np
from scipy import spatial

from arnica.utils.sample_curve_from_cloud import (sort_points_by_dist,
                                                  sample_arrays_by_dist_interval,
                                                  sample_points_from_cloud,
                                                  get_neighbor)

def test_get_neighbor():
    """ Testing get_neighbor """

    points_coor = np.array([[0., 0.],
                            [0., 3.],
                            [1., 0.],
                            [1., 3.]])
    kdtree = spatial.cKDTree(points_coor) #pylint: disable=not-callable
    list_points = [0, 2]
    point = points_coor[2]

    index, dist = get_neighbor(kdtree, point, list_points)

    index_target = 3
    assert index == index_target
    dist_target = 3.0
    assert dist == dist_target

def test_sort_points_by_dist():
    """ Testing sort_points_by_dist """

    starting_pt = np.array([-1., 0.])
    points_coor = np.array([[2., 1.],
                            [1., 2.],
                            [0., 0.],
                            [2., 0.],
                            [2., 2.]])

    ordered_indexes, ordered_dists = sort_points_by_dist(points_coor, starting_pt)

    indexes_target = np.array([2, 3, 0, 4, 1])
    np.testing.assert_array_equal(ordered_indexes, indexes_target)

    dists_target = np.array([0., 2., 1., 1., 1.])
    np.testing.assert_array_equal(ordered_dists, dists_target)

def test_sample_arrays_by_dist_interval():
    """ Testing sample_arrays_by_dist_interval """

    dists = np.zeros(7)
    dists[1:] = 1.
    sample_res = 1.5
    args = np.arange(7.)

    dists_sampled, (args_sampled,) = sample_arrays_by_dist_interval(dists,
                                                                    sample_res,
                                                                    args)

    args_sampled_target = np.array([0., 2., 3., 6.])
    np.testing.assert_array_equal(args_sampled, args_sampled_target)
    dists_sampled_target = np.array([0., 1., 1., 1.])
    np.testing.assert_array_equal(dists_sampled, dists_sampled_target)

def test_sample_points_from_cloud():
    """ Testing sample_points_from_cloud """

    idx_in = np.arange(16)
    np.random.shuffle(idx_in)
    abs_curv = np.linspace(0., 2*np.pi, len(idx_in) + 1)[:-1]
    abs_curv = abs_curv[idx_in]
    points_coor = np.stack((np.cos(abs_curv),
                            np.sin(abs_curv)),
                           axis=1)
    starting_pt = np.array([1.2, 0.])
    n_samples = 8

    out_points = sample_points_from_cloud(points_coor, starting_pt, n_samples=n_samples)

    abs_curv = np.linspace(0., 2*np.pi, n_samples + 1)[:-1]
    abs_curv[-1] = 15 / 16 * 2 * np.pi
    out_points_target = np.stack((np.cos(abs_curv),
                                  np.sin(abs_curv)),
                                 axis=1)
    np.testing.assert_allclose(out_points, out_points_target)

    idx_in = np.arange(100)
    np.random.shuffle(idx_in)
    points_coor = np.zeros((100, 2))
    points_coor[:, 0] = np.linspace(0., 99., 100)[idx_in]
    out_points = sample_points_from_cloud(points_coor, starting_pt)
    np.testing.assert_allclose(len(out_points), 80, atol=2, rtol=2)



""" Module testing Shell, AxiShel et CartShell methods """
import pytest
import os
import numpy as np
import h5py
from arnica.utils.axishell_2 import AxiShell
from arnica.utils.cartshell_3 import CartShell

def test_shell(datadir):
    """ Test of Shell methods by the mean of an AxiShell """

    ctrl_pts_x = (0.0, 0.2)
    ctrl_pts_r = (0.12, 0.12)
    shell = AxiShell(11, 11, 40.0, ctrl_pts_x, ctrl_pts_r)

    #Testing add_curviwidth
    width_profile = ((0.0, -0.04),
                     (0.2, -0.04),
                     (0.4, -0.07),
                     (0.6, -0.04),
                     (0.8, -0.04))
    shell.add_curviwidth('width_profile', width_profile)
    target_val = [-0.04, -0.04, -0.04, -0.055, -0.07, -0.055, -0.04, -0.04,
                  -0.04, -0.04, -0.04]
    np.testing.assert_allclose(shell.width_matrix['width_profile'][0], target_val)

    #Testing bake_millefeuille
    cake = shell.bake_millefeuille('width_profile', 3)
    target_val = [[0.00666667, 0.01333333, 0.00666667],
                  [0.00666667, 0.01333333, 0.00666667],
                  [0.00666667, 0.01333333, 0.00666667],
                  [0.00916667, 0.01833333, 0.00916667],
                  [0.01166667, 0.02333333, 0.01166667],
                  [0.00916667, 0.01833333, 0.00916667],
                  [0.00666667, 0.01333333, 0.00666667],
                  [0.00666667, 0.01333333, 0.00666667],
                  [0.00666667, 0.01333333, 0.00666667],
                  [0.00666667, 0.01333333, 0.00666667],
                  [0.00666667, 0.01333333, 0.00666667]]
    np.testing.assert_allclose(cake['dz'][0], target_val, rtol=1e-5)

    #Testing average_on_shell_over_dirs
    np.random.seed(1)
    target_field = np.random.random((3, shell.shape[0], shell.shape[1])) #pylint: disable=no-member
    average = shell.average_on_shell_over_dirs(target_field, ['v', 'time'])
    target_val = [3.71048106e-05 ,7.33521695e-05, 8.52748890e-05, 8.09592620e-05,
                  8.20054358e-05, 6.79632056e-05, 7.66314353e-05, 7.87668173e-05,
                  8.60460952e-05, 7.16661617e-05, 4.32831721e-05]
    np.testing.assert_allclose(average, target_val, rtol=1e-5)
    average = shell.average_on_shell_over_dirs(target_field, ['v', 'time'], scale=False)
    target_val = [0.47320114, 0.4699923 , 0.55265719, 0.53329244,
                  0.52278534, 0.44020879, 0.51550363, 0.52670343,
                  0.55136021, 0.46839039, 0.57088887]
    np.testing.assert_allclose(average, target_val, rtol=1e-5)
    
    with pytest.raises(ValueError):
        shell.average_on_shell_over_dirs(target_field, ['v', 'k'])

    point_cloud = np.array([[0.1, 0.12, 0.0]])
    tolerance = 0.04
    mask = shell.set_mask_on_shell(point_cloud, tolerance)
    target_field = np.array([[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                             [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
                             [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
                             [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
                             [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
                             [1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1],
                             [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
                             [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
                             [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
                             [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
                             [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]])
    np.testing.assert_array_equal(mask, target_field)

    case = datadir
    path_actu = os.getcwd()
    os.chdir(case)

    shell.plot("fig.png")
    
    shell.dump_shell("shell")
    size_target = shell.shape[0] * shell.shape[1]
    with h5py.File("shell.h5", 'r') as fin:
        assert fin['mesh/x'][()].size == size_target
        assert fin['mesh/y'][()].size == size_target
        assert fin['mesh/z'][()].size == size_target
        assert fin['variables/r'][()].size == size_target
        assert fin['variables/theta'][()].size == size_target
        assert fin['variables/n_x'][()].size == size_target
        assert fin['variables/n_y'][()].size == size_target
        assert fin['variables/n_z'][()].size == size_target
        assert fin['variables/n_r'][()].size == size_target
        assert fin['variables/du'][()].size == size_target
        assert fin['variables/dv'][()].size == size_target
        assert fin['variables/dwu'][()].size == size_target
        assert fin['variables/dwv'][()].size == size_target
        assert fin['variables/surf'][()].size == size_target

    os.chdir(path_actu)



"""Test plot density mesh"""
import numpy as np
from arnica.utils import plot_density_mesh as pdm
import pytest

@pytest.mark.skip(reason="Need to investigate? How to test functions in function")
def test_get_bins():

	x_coord = [0, .5,  1,  0, .5,  1,  0,  1]
	y_coord = [0,  0,  0,  1,  1,  1, .5, .5]
	z_coord = [0,  0,  0,  0,  0,  0,  0,  0]

	#Testing _get_bins

	a,b = pdm.get_bins(x_coord, y_coord)
	assert((a,b) == (500,500))

def test_scatter_plot_mesh():
	x_coord = [0, .5,  1,  0, .5,  1,  0,  1]
	y_coord = [0,  0,  0,  1,  1,  1, .5, .5]
	z_coord = [0,  0,  0,  0,  0,  0,  0,  0]
	pdm.scatter_plot_mesh(x_coord, y_coord, z_coord, axisym=False, show=False)
	pdm.scatter_plot_mesh(x_coord, y_coord, z_coord, axisym=True, show=False)




"""
Functional Testing of the p1 radiation solver
"""
from copy import deepcopy
import numpy as np
import pytest

from arnica.solvers_2d.radiation import main_p1_solver
# from tests_scripts.parameters_tests import (DATA_DIR,
#                                             OUT_DIR)
from tests.parameters_tests import (DATA_DIR, OUT_DIR)

GENERATE_TEST_DATA = False
#GENERATE_TEST_DATA = True


@pytest.fixture
def test_case_cyl():
    """
    Computational setup corresponding to the test case used for validation:
        >   2 concentric cylinder with grey gas
        >   (Modest's book, 2nd edition, p. 477)

    """
    t_west = 500.
    t_east = 1500.

    mesh_cyl = {"kind": 'cyl',
                "r_min": 1.,
                "r_max": 2.,
                "theta_min": 0.,
                "theta_max": 2. * np.pi,
                "n_pts_rad": 5
                }

    # Boundary conditions
    # Default values
    boundary = {"North": {"type": "Periodic", 'Values': {}},
                "West": {"type": "Periodic", 'Values': {}},
                "South": {"type": "Periodic", 'Values': {}},
                "East": {"type": "Periodic", 'Values': {}}}

    # Overwrite WE BCs with Robin conditions
    boundary["West"]["type"] = "Robin"
    boundary["West"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 500.}

    boundary["East"]["type"] = "Robin"
    boundary["East"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 1500.}

    # Input fields: Temperature, pressure, etc.
    field = {"absorption_coefficient": 1.,
             "temperature": [t_west, t_east]}

    return {"boundaries": boundary, "field": field, "mesh": mesh_cyl}


def test_p1_solver_is_running(test_case_cyl):
    """
    Test if the solver is running
    """
    # params = deepcopy(test_case_cyl)
    params = test_case_cyl
    q_inner, _ = main_p1_solver(params)
    assert q_inner


def test_p1_solver_q_inner_is_correct(test_case_cyl):
    """
    Test if the solver is giving good results for q_inner at several optical
    lengths
    """
    params = deepcopy(test_case_cyl)
    abs_coeff = np.linspace(1, 6, 10)
    data_path = "%s/p1/q_inner_test_data.npz" %DATA_DIR

    q_inner_res = []
    for kappa in abs_coeff:
        params['field']['absorption_coefficient'] = kappa
        q_inner, _ = main_p1_solver(params)
        q_inner_res.append(q_inner)

    if GENERATE_TEST_DATA:
        np.savez(data_path, q_inner=q_inner_res)
    else:
        data = np.load(data_path)
        q_inner_saved = data["q_inner"]
        print(q_inner_saved)
        print(q_inner_res)
        
        assert np.allclose(q_inner_saved, np.array(q_inner_res), atol=1e-5)


def test_p1_solver_fields_are_correct(test_case_cyl):
    """
    Test if the solver is giving good results for all the fields!
    """
    params = deepcopy(test_case_cyl)
    # data_path = "%s/p1/fields_test_data.npz" %DATA_DIR
    data_path = DATA_DIR +  "/p1/fields_%s_test_data.npz"
    params['field']['absorption_coefficient'] = 3.
    _, results = main_p1_solver(params)

    if GENERATE_TEST_DATA:
        for key in results.keys():
            try:
                np_data = results[key].toarray()
            except AttributeError:
                np_data = results[key]

            np.savez(data_path % key, data=np_data)
    else:
        for key in results.keys():
            data = np.load(data_path % key)
            res_saved = data['data']
            res = np.array(results[key])
            assert np.allclose(res, res_saved)




import numpy as np
from arnica.utils.showy import showy


def test_showy():
    """Damo of a simplified set of graphs with showy"""
    data = dict()
    data["time"] = np.linspace(0, 0.1, num=256)

    freq = 10.
    for freq in np.linspace(10, 20, num=9):
        data["sine_" + str(freq)] = np.cos(data["time"] * freq * 2 * np.pi)

    # Creating a template
    template = {
        "title": "Example",
        "graphs": [{
            "curves": [{"var": "sine_*"}],
            "x_var": "time",
            "y_label": "Sine [mol/m³/s]",
            "x_label": "Time [s]",
            "title": "Sinus of frquency *"
        }],
        "figure_structure": [3, 3],
        "figure_dpi": 92.6
    }

    showy(template, data, data_c=None, show=False)



import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

""" unit tests for arnica.utils

OST : I heard it through the grapevine, by Leo. Moracchioli
"""

import numpy as np
from arnica.utils.vector_actions import (
    renormalize,
    angle_btw_vects,
    yz_to_theta,
    rtheta2yz,
    rotate_vect_around_x,
    rotate_vect_around_axis,
    dilate_vect_around_x,
    cart_to_cyl,
    clip_by_bounds,
    cyl_to_cart,
    vect_to_quat,
    make_radial_vect,
    mask_cloud,
)

from arnica.utils.directed_projection import (
    projection_kdtree,
    compute_dists,
    project_points,
)
from arnica.utils.cloud2cloud import cloud2cloud
from arnica.utils.nparray2xmf import NpArray2Xmf
from arnica.utils.plot_quad2tri import (
    get_connectivity,
    quad2tri,
    plot_quad2tri,
)

SMALL = 3e-16
DIM1 = 5
DIM2 = 7

BINS = 100


def deviation(fieldx_in, fieldy_in):
    """ plot deviation """

    fieldx = np.ravel(fieldx_in)
    fieldy = np.ravel(fieldy_in)
    corr_norm = np.linalg.norm(fieldx - fieldy, ord=2)
    corr_norm /= np.linalg.norm((fieldx + fieldy) / 2., ord=2)

    plt.cla()
    plt.hist2d(fieldx, fieldy, BINS, norm=LogNorm())
    line45 = np.linspace(np.amin(np.concatenate((fieldx, fieldy), axis=0)),
                         np.amax(np.concatenate((fieldx, fieldy), axis=0)))
    plt.plot(line45, line45, linewidth=2.0, ls='--', c='k')
    plt.xlim(np.amin(fieldx), np.amax(fieldx))
    plt.ylim(np.amin(fieldx), np.amax(fieldx))
    plt.xlabel('File 1', fontsize=16)
    plt.ylabel('File 2', fontsize=16)
    plt.tick_params(axis='x', labelsize=16)
    plt.tick_params(axis='y', labelsize=16)
    textstr = ("L2-norm error:\n" + str(corr_norm))
    plt.gcf().text(0.02,
                   0.02,
                   textstr,
                   fontsize=18,
                   bbox={'facecolor': 'red',
                         'alpha': 0.2})
    plt.rc('font', size=16)
    plt.tight_layout()
    plt.show()


def squared_func(xxx, yyy, zzz):
    """ test function """
    return (3 * xxx ** 2 + 5 * yyy ** 2 + 7 * zzz * 2) / (
            3 ** 2 * 5 ** 2 * 7 ** 2)


def test_nparray2xmf():
    """ test of nparray dump """

    x_coor = np.linspace(0, 1., DIM2 * 10)
    y_coor = np.linspace(0, 1., DIM2 * 10)
    z_coor = np.linspace(0, 1., DIM2 * 10)

    source_xyz = np.stack(np.meshgrid(x_coor, y_coor, z_coor), axis=-1)

    xmf3d = NpArray2Xmf("dummy.h5")
    xmf3d.create_grid(np.take(source_xyz, 0, axis=-1),
                      np.take(source_xyz, 1, axis=-1),
                      np.take(source_xyz, 2, axis=-1))

    xmf3d.add_field(squared_func(np.take(source_xyz, 0, axis=-1),
                                 np.take(source_xyz, 1, axis=-1),
                                 np.take(source_xyz, 2, axis=-1)),
                    "dummy")


def test_cloud2cloud():
    """
    small test of cloud2cloud projection
    """
    size = 1000
    x_coor = np.linspace(0, 1., DIM2 * 10)
    y_coor = np.linspace(0, 1., DIM2 * 10)
    z_coor = np.linspace(0, 1., DIM2 * 10)

    source_xyz = np.stack(np.meshgrid(x_coor, y_coor, z_coor),
                          axis=-1)
    target_xyz = np.stack((np.random.rand(DIM1 * size) * 1.,
                           np.random.rand(DIM1 * size) * 1.,
                           np.random.rand(DIM1 * size) * 1.),
                          axis=-1)

    source_val = {}
    source_val["v"] = squared_func(np.take(source_xyz, 0, axis=-1),
                                   np.take(source_xyz, 1, axis=-1),
                                   np.take(source_xyz, 2, axis=-1), )

    target_ref = squared_func(target_xyz[:, 0],
                              target_xyz[:, 1],
                              target_xyz[:, 2], )

    target_estimate5 = cloud2cloud(source_xyz,
                                   source_val,
                                   target_xyz,
                                   stencil=5)

    target_estimate1 = cloud2cloud(source_xyz,
                                   source_val,
                                   target_xyz,
                                   stencil=1,
                                   limitsource=100000)
    assert np.all(np.abs(target_ref - target_estimate1["v"]) < 1e-4)
    assert np.all(np.abs(target_ref - target_estimate5["v"]) < 1e-5)

def test_clip_by_bounds():
    """ test the clip of a cloud of points by bounds """

    point_cloud = np.array([[0., 0., 1.],
                            [1., np.cos(np.pi/4), np.sin(np.pi/4)],
                            [2., 1., 0.]])
    bounds_dict = {"x": (0.5, 3.0), "y": (0.0, 0.8)}
    exp_out = clip_by_bounds(point_cloud, bounds_dict, keep="out")
    exp_target = np.array([[0., 0., 1.],
                           [2., 1., 0.]])

    exp_out = clip_by_bounds(point_cloud, bounds_dict, keep="in")
    exp_target = np.array([[1., np.cos(np.pi/4), np.sin(np.pi/4)]])

    np.testing.assert_array_equal(exp_out, exp_target)

def test_projection_kdtree():
    r"""
    small test of directed projection

                  xyz(0,1,0)  dir(1,-1,0)
                x
                 \
                 _\|


                ^
                |
  ______________o______x________
  s0     s1     s2     s3     s4
               /          xyz(1,0,0)
xyz(0,0,0)  nml (0,1,0)
    """

    test_points_xyz = np.zeros((DIM2, 3))
    test_points_xyz[:, 1] = 1.0
    test_points_dir = np.zeros((DIM2, 3))
    test_points_dir[:, 0] = 1.0
    test_points_dir[:, 1] = -1.0
    test_points_dir = renormalize(test_points_dir)

    test_surf_xyz = np.zeros((DIM1, 3))
    test_surf_xyz[:, 0] = np.linspace(-2, 2, DIM1)
    test_surf_nml = np.zeros((DIM1, 3))
    test_surf_nml[:, 1] = 1.0

    neighb = 3

    projection = projection_kdtree(
        test_points_xyz,
        test_points_dir,
        test_surf_xyz,
        test_surf_nml,
        neigbors=neighb,
    )

    exp_xyz = np.zeros((DIM2, 3))
    exp_indexes = np.repeat(
        np.array([2, 3, 1])[np.newaxis, :],
        DIM2,
        axis=0)
    exp_cyl_dists = np.repeat(
        np.array([0.0, np.sqrt(2.0) * 0.5, np.sqrt(2.0) * 0.5])[np.newaxis, :],
        DIM2,
        axis=0)

    np.testing.assert_allclose(projection.moved_pts, exp_xyz)
    np.testing.assert_equal(projection.indexes, exp_indexes)
    np.testing.assert_allclose(projection.cyl_dists, exp_cyl_dists, atol=1e-6)
    # assert out_xyz.shape == projection.moved_pts.shape
    # assert np.all(out_xyz == projection.moved_pts)
    # assert out_indexes.shape == projection.indexes.shape
    # assert np.all(out_indexes == projection.indexes)
    # assert out_cyl_dists.shape == exp_out_cyl_dists.shape
    # assert np.all(np.abs(exp_out_cyl_dists - out_cyl_dists) < SMALL)

    project=False
    projection = projection_kdtree(
        test_points_xyz,
        test_points_dir,
        test_surf_xyz,
        test_surf_nml,
        neigbors=neighb,
        project=project)

    exp_xyz = np.tile([0., 1., 0.], (DIM2, 1))
    exp_indexes = np.tile([2, 3, 1], (DIM2, 1))
    exp_cyl_dists = np.tile([0.70710678, 0., 1.4142135], (DIM2, 1))

    np.testing.assert_allclose(projection.moved_pts, exp_xyz)
    np.testing.assert_equal(projection.indexes, exp_indexes)
    np.testing.assert_allclose(projection.cyl_dists, exp_cyl_dists, atol=1e-6)

def test_compute_dists():
    """ test of computation of cylindrical distances """

    test_points_xyz = np.zeros((DIM2, 3))
    test_points_dir = np.zeros((DIM2, 3))
    test_points_dir[:, 0] = 1.0
    test_points_dir[:, 1] = -1.0
    test_points_dir = renormalize(test_points_dir)

    neighb = 3
    test_surf_xyz = np.zeros((DIM2, neighb, 3))
    test_surf_xyz[:, 1, 0] = 1.
    test_surf_xyz[:, 2, 0] = -1.
    test_surf_nml = np.zeros((DIM2, neighb, 3))
    test_surf_nml[:, :, 1] = 1.
    
    tol = 1000.

    _, out_cyl_dists = compute_dists(
        test_points_xyz,
        test_points_dir,
        test_surf_xyz,
        test_surf_nml,
        tol,
    )
    exp_out_cyl_dists = np.array([[0., 0.70710678, 0.70710678],
                                  [0., 0.70710678, 0.70710678],
                                  [0., 0.70710678, 0.70710678],
                                  [0., 0.70710678, 0.70710678],
                                  [0., 0.70710678, 0.70710678],
                                  [0., 0.70710678, 0.70710678],
                                  [0., 0.70710678, 0.70710678]])

    assert out_cyl_dists.shape == exp_out_cyl_dists.shape
    np.testing.assert_allclose(out_cyl_dists, exp_out_cyl_dists)


def test_project_points():
    """ test the projection of points on a plan defined by a point and a normal """

    test_points_xyz = np.zeros((DIM2, 3))
    test_points_xyz[:, 1] = 1.
    test_points_dir = np.zeros((DIM2, 3))
    test_points_dir[:, 1] = -1.
    test_surf_xyz = np.zeros((DIM2, DIM1, 3))
    test_surf_xyz[:,:,0] = np.linspace(-2., 2., DIM1)

    proj = project_points(test_points_xyz, test_points_dir, test_surf_xyz)
    
    exp_out_proj_pts = np.tile(np.array([0., 0., 0.]), (DIM2, DIM1, 1))
    exp_out_axi_dists = np.tile(np.array([1., 1., 1., 1., 1.]), (DIM2, 1))
    exp_out_rad_dists = np.tile(np.array([2., 1., 0., 1., 2.]), (DIM2, 1))
    np.testing.assert_array_equal(exp_out_proj_pts, proj.proj_pts)
    np.testing.assert_array_equal(exp_out_axi_dists, proj.axi_dists)
    np.testing.assert_array_equal(exp_out_rad_dists, proj.rad_dists)

def test_renormalize():
    """ test renormalize in 1D an 2D arrays of 3D components"""
    in_x = np.ones((DIM1, 3))
    out_x = in_x / np.sqrt(3)
    assert np.all(renormalize(in_x) == out_x)
    in_x = np.ones((DIM1, DIM2, 3))
    out_x = in_x / np.sqrt(3)
    assert np.all(renormalize(in_x) == out_x)


def test_angle_btw_vects():
    """ test angle btw vects in 1D an 2D arrays of 3D components"""
    # todo : failing test to fix
    in_x1 = np.zeros((DIM1, 3))
    in_x1[:, 2] = 1.0
    in_x2 = np.zeros((DIM1, 3))
    in_x2[:, 1] = 1.0
    out = angle_btw_vects(in_x1, in_x2, convert_to_degree=True)

    assert np.all(out == np.ones((DIM1)) * 90.)

    in_x1 = np.zeros((DIM1, DIM2, 3))
    in_x1[:, :, 2] = 1.0
    in_x2 = np.zeros((DIM1, DIM2, 3))
    in_x2[:, :, 1] = 1.0
    out = angle_btw_vects(in_x1, in_x2, convert_to_degree=True)
    assert np.all(out == np.ones((DIM1, DIM2)) * 90.)


def test_yz_to_theta():
    """ test convertion to theta
           0pi=0deg
              Y
              ^
              |
              |
-0.5pi=-90deg     o------>Z   0.5pi=90deg
    """
    in_x = np.zeros((DIM1, 3))
    in_x[:, 2] = 1.0
    out = yz_to_theta(in_x)
    assert np.all(out == np.ones((DIM1)) * np.pi * 0.5)

    in_x = np.zeros((DIM1, 3))
    in_x[:, 1] = 1.0
    out = yz_to_theta(in_x)
    assert np.all(out == np.ones((DIM1)) * 0.)

    in_x = np.zeros((DIM1, DIM2, 3))
    in_x[:, :, 2] = -1.0
    out = yz_to_theta(in_x)
    assert np.all(out == np.ones((DIM1, DIM2)) * -0.5 * np.pi)

def test_rtheta2yz():
    """ test conversion of r, theta to y, z-axis """

    in_r = np.ones(DIM1) * 2.
    in_theta = np.ones(DIM1) * np.pi / 4.
    out = rtheta2yz(in_r, in_theta)
    np.testing.assert_allclose(out[0], np.ones(DIM1) * 1.414213562373095)
    np.testing.assert_allclose(out[1], np.ones(DIM1) * 1.414213562373095)

    in_r = np.ones((DIM1, DIM2)) * 2.
    in_theta = np.ones((DIM1, DIM2)) * np.pi * 3 / 4
    out = rtheta2yz(in_r, in_theta)
    np.testing.assert_allclose(out[0], np.ones((DIM1, DIM2)) * -1.414213562373095)
    np.testing.assert_allclose(out[1], np.ones((DIM1, DIM2)) * 1.414213562373095)

def test_rotate_vect_around_x():
    """ test the rotation of vectors """
    in_x = np.zeros((DIM1, 3))
    in_x[:, 2] = 1.0
    out = rotate_vect_around_x(in_x, 90.0)
    exp_out = np.zeros((DIM1, 3))
    exp_out[:, 1] = -1.0
    assert np.all(np.abs(out - exp_out) < SMALL)

    in_x = np.zeros((DIM1, DIM2, 3))
    in_x[:, :, 2] = 1.0
    out = rotate_vect_around_x(in_x, 90.0)
    exp_out = np.zeros((DIM1, DIM2, 3))
    exp_out[:, :, 1] = -1.0
    assert out.shape == exp_out.shape
    assert np.all(np.abs(out - exp_out) < SMALL)

def test_rotate_vect_around_axis():
    """ test of rotation of vectors around several axis """

    in_x = np.zeros((DIM1, 3))
    in_x[:, 2] = 1.0
    out = rotate_vect_around_axis(in_x, ([1, 0, 0], 90.0))
    exp_out = np.zeros((DIM1, 3))
    exp_out[:, 1] = -1.0
    assert np.all(np.abs(out - exp_out) < SMALL)

    out = rotate_vect_around_axis(in_x,
                                  ([0, 1, 0], 45.),
                                  ([4., 0., 2.1], 64.))
    exp_out = np.tile([0.55922829, -0.79579036, 0.232339], (DIM1, 1))
    np.testing.assert_allclose(out, exp_out)

def test_dilate_vect_around_x():
    """ test the dilatation of vectors around x """

    azimuth = np.array([-45., 0., 45.])
    np_ar_vect = np.array([[0., np.cos(np.pi/4), -np.sin(np.pi/4)],
                           [0., 1., 0.],
                           [0., np.cos(np.pi/4), np.sin(np.pi/4)]])
    exp_out = dilate_vect_around_x(azimuth, np_ar_vect)
    exp_target = np.array([[0., -1., 0.],
                           [0., 1., 0.],
                           [0., -1., 0.]])

    np.testing.assert_allclose(exp_out, exp_target, atol=1e-8)

def test_cart_to_cyl():
    """ test the conversion from xyz-system to cyl-system """

    vects_xyz = np.array([[0., 0., 1.],
                          [1., np.cos(np.pi/4), np.sin(np.pi/4)],
                          [2., 1., 0.]])
    exp_out = cart_to_cyl(vects_xyz)
    exp_target = np.array([[0., 1., np.pi/2],
                           [1., 1., np.pi/4],
                           [2., 1., 0.]])

    np.testing.assert_allclose(exp_out, exp_target)

def test_cyl_to_cart():
    """ test the conversion from cyl-system to xyz-system """

    vects_cyl = np.array([[0., 1., np.pi/2],
                          [1., 1., np.pi/4],
                          [2., 1., 0.]])
    exp_out = cyl_to_cart(vects_cyl)
    exp_target = np.array([[0., 0., 1.],
                           [1., np.cos(np.pi/4), np.sin(np.pi/4)],
                           [2., 1., 0.]])

    np.testing.assert_allclose(exp_out, exp_target, atol=1e-8)

def test_clip_by_bounds():
    """ test of the clip of field by bounds dict """

    points_coord = np.arange(DIM2 * 3.).reshape((DIM2, 3))
    bounds = {"x": (3., 13.)}
    out = clip_by_bounds(points_coord, bounds, keep="in", return_mask=False)
    out_target = [[3.,   4.,  5.],
                  [6.,   7.,  8.],
                  [9.,  10., 11.],
                  [12., 13., 14.]]
    np.testing.assert_array_equal(out, out_target)

    bounds = {"x": (3., 13.),
              "z": (6., 1.e6)}
    out = clip_by_bounds(points_coord, bounds, keep="out", return_mask=True)
    out_target = np.array([1, 1, 0, 0, 0, 1, 1], dtype=bool)
    np.testing.assert_array_equal(out, out_target)

def test_vect_to_quat():
    """ test the building of a quaternion from two vects """

    vect_t = np.array([0., 1., 1.])
    vect_s = np.array([0., 0., 2.])
    exp_out = vect_to_quat(vect_t, vect_s).as_rotvec()
    exp_target = np.array([-np.pi / 4, 0., 0.])

    np.testing.assert_allclose(exp_out, exp_target, atol=1e-8)

    vect_t = np.array([[0., 1., 1.],
                       [1., 2., 0.]])
    vect_s = np.array([[0., 0., 2.],
                       [0., 2., 1.]])
    exp_out = vect_to_quat(vect_t, vect_s).as_rotvec()
    exp_target = np.array([[-0.78539816, 0., 0.],
                           [-0.42900074, 0.21450037, -0.42900074]])

    np.testing.assert_allclose(exp_out, exp_target, atol=1e-8)

def test_make_radial_vect():
    """ test function making vectors radial """

    coord = np.zeros((DIM1, 3))
    coord[:, 1] = 1.
    coord[:, 2] = -1.
    vects = np.zeros((DIM1, 3))
    vects[:, 0] = 1.
    vects[:, 1] = 1.
    vects = renormalize(vects)

    out = make_radial_vect(coord, vects)
    np.testing.assert_allclose(out, vects, atol=1e-6)

    coord[:, 1] = 6.
    out = make_radial_vect(coord, vects)
    out_target = [[0.70710678, 0.69748583, -0.11624764],
                  [0.70710678, 0.69748583, -0.11624764],
                  [0.70710678, 0.69748583, -0.11624764],
                  [0.70710678, 0.69748583, -0.11624764],
                  [0.70710678, 0.69748583, -0.11624764]]
    np.testing.assert_allclose(out, out_target, atol=1e-6)

def test_mask_cloud():
    """ test vector masking """
    in_x = np.stack((np.linspace(-2, 2, DIM1),
                     np.linspace(-2, 2, DIM1),
                     np.linspace(-2, 2, DIM1)),
                    axis=0)
    for i, axe in enumerate(["x", "y", "z"]):
        out = mask_cloud(in_x,
                         axis=axe,
                         support=(-1.0, 1.0))
        exp_out = (in_x[:, i] >= -1.0) * (in_x[:, i] < 1.0)
        assert np.all(out == exp_out)

    in_x = np.repeat(in_x[:, np.newaxis, :],
                     DIM2,
                     axis=1)
    for i, axe in enumerate(["x", "y", "z"]):
        out = mask_cloud(in_x,
                         axis=axe,
                         support=(-1.0, 1.0))
        exp_out = (in_x[:, :, i] >= -1.0) * (in_x[:, :, i] < 1.0)
        assert out.shape == exp_out.shape
        assert np.all(out == exp_out)

    in_x = np.stack((np.zeros(DIM1),
                     np.ones(DIM1),
                     np.linspace(-2, 2, DIM1)),
                    axis=0)
    out = mask_cloud(in_x,
                     axis="r",
                     support=(0, 2.0))
    exp_out = np.hypot(np.take(in_x, 1, axis=-1),
                       np.take(in_x, 2, axis=-1)) < 2.0
    assert np.all(out == exp_out)

    out = mask_cloud(in_x,
                     axis="theta",
                     support=(-20.0, 20))
    theta = np.rad2deg(yz_to_theta(in_x))
    exp_out = (theta >= -20.0) * (theta < 20.0)
    assert np.all(out == exp_out)

    exception_reached = False
    try:
        mask_cloud(in_x,
                   axis="dummy",
                   support=(-20.0, 20))
    except IOError:
        exception_reached = True

    assert exception_reached

def test_get_connectivity():
    """ test get_connectivity """

    i = 1
    j = 2
    shape = (4, 4)

    connectivity_target = [[6, 7, 21], [7, 11, 21], [11, 10, 21], [10, 6, 21]]
    np.testing.assert_array_equal(get_connectivity(i, j, shape), connectivity_target)

def test_quad2tri():
    """ test of quad2tri """

    x, y = np.meshgrid(np.linspace(0., 1., 3),
                       np.linspace(0., 1., 5))
    grid = np.stack((y, x), axis=-1)

    field = np.ones((5,3))
    field[:, 0] = 2.

    triangulation, field = quad2tri(grid, field)

    field_tgt = [2., 1., 1., 2., 1., 1., 2., 1., 1.,
                 2., 1., 1., 2., 1., 1., 1.5, 1., 1.5,
                 1., 1.5, 1., 1.5, 1.]
    np.testing.assert_array_equal(field, field_tgt)

def test_plot_quad2tri():
    """ test of plotting of quad2tri """

    x, y = np.meshgrid(np.linspace(0., 1., 3),
                       np.linspace(0., 1., 5))
    grid = np.stack((y, x), axis=-1)
    field = np.ones((5,3))
    field[:, 0] = 2.
    title = "Test_plot_quad2tri"

    plt_out = plot_quad2tri(grid, title, field)
    assert isinstance(plt_out, type(plt))
    plt_out = plot_quad2tri(grid, title, field, shading=False)
    assert isinstance(plt_out, type(plt))



""" module to test de core_fd parts of Arnica """

from copy import deepcopy

import numpy as np
import pytest

import arnica.utils.mesh_tool as msh
from arnica.solvers_2d.core_fd import Metrics2d
from arnica.solvers_2d.boundary import Boundary2d


@pytest.fixture
def setup_test_case_cyl():
    T_west = 500.
    T_east = 1500.

    MESH_CYL = {"kind": 'cyl',
                "r_min": 1.,
                "r_max": 2.,
                "theta_min": 0.,
                "theta_max": 2. * np.pi,
                "n_pts_rad": 15,
                "n_pts_tgt": 15
                }

    # Boundary conditions
    # Default values
    boundary = {"North": {"type": "Periodic", 'Values': {}},
                "West": {"type": "Periodic", 'Values': {}},
                "South": {"type": "Periodic", 'Values': {}},
                "East": {"type": "Periodic", 'Values': {}}}

    boundary["West"]["type"] = "Robin"
    boundary["West"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 500.}

    boundary["East"]["type"] = "Robin"
    boundary["East"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 1500.}

    # Input fields: Temperature, pressure, etc.
    field = {"absorption_coefficient": 1.,
             "temperature": [T_west, T_east]}

    # Create parameter dictionary
    params = {}
    params["boundaries"] = boundary
    params["field"] = field

    # Setup simulation with cylindrical mesh
    params["mesh"] = MESH_CYL

    return params


@pytest.fixture
def setup_test_case_cyl_diff_res_theta():
    T_west = 500.
    T_east = 1500.

    MESH_CYL = {"kind": 'cyl',
                "r_min": 1.,
                "r_max": 2.,
                "theta_min": 0.,
                "theta_max": 2. * np.pi,
                "n_pts_rad": 15
                }

    # Boundary conditions
    # Default values
    BOUNDARY = {"North": {"type": "Periodic", 'Values': {}},
                "West": {"type": "Periodic", 'Values': {}},
                "South": {"type": "Periodic", 'Values': {}},
                "East": {"type": "Periodic", 'Values': {}}}

    BOUNDARY["West"]["type"] = "Robin"
    BOUNDARY["West"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 500.}

    BOUNDARY["East"]["type"] = "Robin"
    BOUNDARY["East"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 1500.}

    # Input fields: Temperature, pressure, etc.
    FIELD = {"absorption_coefficient": 1.,
             "temperature": [T_west, T_east]}

    # Create parameter dictionary
    PARAMS = {}
    PARAMS["boundaries"] = BOUNDARY
    PARAMS["field"] = FIELD

    # Setup simulation with cylindrical mesh
    PARAMS["mesh"] = MESH_CYL

    return PARAMS


@pytest.fixture
def setup_test_case_rect_diff_res_x_y():
    T_west = 500.
    T_east = 1500.

    # Mesh definition
    MESH_RECT = {"kind": 'rect',
                 "x_max": 1.,
                 "y_max": 1.,
                 "x_res": 5.,
                 "y_res": 4.,
                 }

    # Boundary conditions
    # Default values
    BOUNDARY = {"North": {"type": "Periodic", 'Values': {}},
                "West": {"type": "Periodic", 'Values': {}},
                "South": {"type": "Periodic", 'Values': {}},
                "East": {"type": "Periodic", 'Values': {}}}

    BOUNDARY["West"]["type"] = "Robin"
    BOUNDARY["West"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 500.}

    BOUNDARY["East"]["type"] = "Robin"
    BOUNDARY["East"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 1500.}

    # Input fields: Temperature, pressure, etc.
    FIELD = {"absorption_coefficient": 1.,
             "temperature": [T_west, T_east]}

    # Create parameter dictionary
    PARAMS = {}
    PARAMS["boundaries"] = BOUNDARY
    PARAMS["field"] = FIELD

    # Setup simulation with rectangular mesh
    PARAMS["mesh"] = MESH_RECT

    return PARAMS


@pytest.fixture
def setup_test_case_rect():
    T_west = 500.
    T_east = 1500.

    # Mesh definition
    MESH_RECT = {"kind": 'rect',
                 "x_max": 1.,
                 "y_max": 1.,
                 "x_res": 5.,
                 "y_res": 5.,
                 }

    # Boundary conditions
    # Default values
    BOUNDARY = {"North": {"type": "Periodic", 'Values': {}},
                "West": {"type": "Periodic", 'Values': {}},
                "South": {"type": "Periodic", 'Values': {}},
                "East": {"type": "Periodic", 'Values': {}}}

    BOUNDARY["West"]["type"] = "Robin"
    BOUNDARY["West"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 500.}

    BOUNDARY["East"]["type"] = "Robin"
    BOUNDARY["East"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 1500.}

    # Input fields: Temperature, pressure, etc.
    FIELD = {"absorption_coefficient": 1.,
             "temperature": [T_west, T_east]}

    # Create parameter dictionary
    PARAMS = {}
    PARAMS["boundaries"] = BOUNDARY
    PARAMS["field"] = FIELD

    # Setup simulation with rectangular mesh
    PARAMS["mesh"] = MESH_RECT

    return PARAMS


def test_perio_ns(setup_test_case_cyl):
    """test NS periodicity"""
    dict_test_perio = deepcopy(setup_test_case_cyl)

    dict_test_perio["boundaries"]["North"]["type"] = "Periodic"
    dict_test_perio["boundaries"]["South"]["type"] = "Periodic"
    dict_test_perio["boundaries"]["West"]["type"] = "Robin"
    dict_test_perio["boundaries"]["East"]["type"] = "Robin"

    bnd = Boundary2d(dict_test_perio["boundaries"])

    assert bnd.periodic_ns == True


def test_perio_we(setup_test_case_cyl):
    """test WE periodicity"""
    dict_test_perio = deepcopy(setup_test_case_cyl)

    dict_test_perio["boundaries"]["West"]["type"] = "Periodic"
    dict_test_perio["boundaries"]["East"]["type"] = "Periodic"
    dict_test_perio["boundaries"]["North"]["type"] = "Robin"
    dict_test_perio["boundaries"]["South"]["type"] = "Robin"

    bnd = Boundary2d(dict_test_perio["boundaries"])

    assert bnd.periodic_we == True


def test_no_perio(setup_test_case_cyl):
    """test no periodicity"""
    dict_test_perio = deepcopy(setup_test_case_cyl)

    dict_test_perio["boundaries"]["West"]["type"] = "Neumann"
    dict_test_perio["boundaries"]["East"]["type"] = "Neumann"
    dict_test_perio["boundaries"]["North"]["type"] = "Robin"
    dict_test_perio["boundaries"]["South"]["type"] = "Robin"

    bnd = Boundary2d(dict_test_perio["boundaries"])

    assert bnd.periodic_we == False
    assert bnd.periodic_ns == False


def test_perio_wrong_bc_definition(setup_test_case_cyl):
    """test wrong bc definition"""
    dict_test_perio = deepcopy(setup_test_case_cyl)

    dict_test_perio["boundaries"]["West"]["type"] = "Periodic"
    dict_test_perio["boundaries"]["North"]["type"] = "Periodic"
    dict_test_perio["boundaries"]["East"]["type"] = "Robin"
    dict_test_perio["boundaries"]["South"]["type"] = "Robin"

    with pytest.raises(IOError):
        Boundary2d(dict_test_perio["boundaries"])


def test_perio_error_only_1_perio_defined(setup_test_case_cyl):
    """test wrong bc definition"""
    dict_test_perio = deepcopy(setup_test_case_cyl)

    dict_test_perio["boundaries"]["West"]["type"] = "Periodic"
    dict_test_perio["boundaries"]["North"]["type"] = "Neumann"
    dict_test_perio["boundaries"]["East"]["type"] = "Robin"
    dict_test_perio["boundaries"]["South"]["type"] = "Robin"

    with pytest.raises(IOError):
        Boundary2d(dict_test_perio["boundaries"])


def test_bc_type(setup_test_case_cyl):
    """test wrong bc implemented"""
    implemented_bc = ["Dirichlet", "Neumann", "Robin", "Periodic"]
    dict_test_perio = deepcopy(setup_test_case_cyl)

    dict_test_perio["boundaries"]["West"]["type"] = "Dirichlet"
    dict_test_perio["boundaries"]["North"]["type"] = "Dirichlet"
    dict_test_perio["boundaries"]["East"]["type"] = "Neumann"
    dict_test_perio["boundaries"]["South"]["type"] = "Robin"

    bnd = Boundary2d(dict_test_perio["boundaries"])

    for bc_name in bnd.boundary_params.keys():
        assert bnd.boundary_params[bc_name]["type"] in implemented_bc


@pytest.fixture
def metric_cyl(setup_test_case_cyl):
    """
    Radiation solver (P1 Approximation)

    Parameters :
    ------------
    params : dict of setup parameters

    """
    params = deepcopy(setup_test_case_cyl)
    x_coor, y_coor = msh.get_mesh(params["mesh"])
    print("Shape of x_coor (cyl)")
    print(x_coor.shape)

    bnd = Boundary2d(params["boundaries"])

    print("\n\tmain ---> Compute Metrics")
    metric = Metrics2d(x_coor, y_coor, periodic_ns=bnd.periodic_ns,
                       periodic_we=bnd.periodic_we)

    return metric


@pytest.fixture
def metric_cyl_diff_res_x_y(setup_test_case_cyl_diff_res_theta):
    """
    Radiation solver (P1 Approximation)

    Parameters :
    ------------
    params : dict of setup parameters

    """
    params = deepcopy(setup_test_case_cyl_diff_res_theta)
    x_coor, y_coor = msh.get_mesh(params["mesh"])
    print("Shape of x_coor (cyl)")
    print(x_coor.shape)

    bnd = Boundary2d(params["boundaries"])

    print("\n\tmain ---> Compute Metrics")
    metric = Metrics2d(x_coor, y_coor, periodic_ns=bnd.periodic_ns,
                       periodic_we=bnd.periodic_we)

    return metric


@pytest.fixture
def metric_rect(setup_test_case_rect):
    """
    Radiation solver (P1 Approximation)

    Parameters :
    ------------
    params : dict of setup parameters

    """
    params = deepcopy(setup_test_case_rect)
    x_coor, y_coor = msh.get_mesh(params["mesh"])

    bnd = Boundary2d(params["boundaries"])

    metric = Metrics2d(x_coor,
                       y_coor,
                       periodic_ns=bnd.periodic_ns,
                       periodic_we=bnd.periodic_we)

    return metric


@pytest.fixture
def metric_rect_diff_res_x_y(setup_test_case_rect_diff_res_x_y):
    """
    Radiation solver (P1 Approximation)

    Parameters :
    ------------
    params : dict of setup parameters

    """
    params = deepcopy(setup_test_case_rect_diff_res_x_y)
    x_coor, y_coor = msh.get_mesh(params["mesh"])

    bnd = Boundary2d(params["boundaries"])

    metric = Metrics2d(x_coor,
                       y_coor,
                       periodic_ns=bnd.periodic_ns,
                       periodic_we=bnd.periodic_we)

    return metric


def test_Metrics2d_grad_x_vs_grad_x_csr_rect(metric_rect):
    """test gradx vs gradx_csr for resx=resy with rect mesh"""
    metric_rect.check_compute_matrices()
    grad_x = metric_rect.grad_x_slow.toarray()
    grad_x_csr = metric_rect.grad_x_csr.toarray()
    assert np.allclose(grad_x, grad_x_csr)


def test_Metrics2d_grad_y_vs_grad_y_csr_rect(metric_rect):
    """test grady vs gradx_csr for resx=resy with rect mesh"""
    metric_rect.check_compute_matrices()
    grad_y = metric_rect.grad_y_slow.toarray()
    grad_y_csr = metric_rect.grad_y_csr.toarray()
    assert np.allclose(grad_y, grad_y_csr)


def test_Metrics2d_grad_x_vs_grad_x_csr_rect_diff_res_x_y(
        metric_rect_diff_res_x_y):
    """Check if csr is correctly implemented for gradx in diff res rect"""
    metric_rect_diff_res_x_y.check_compute_matrices()
    grad_x = metric_rect_diff_res_x_y.grad_x_slow.toarray()
    grad_x_csr = metric_rect_diff_res_x_y.grad_x_csr.toarray()
    assert np.allclose(grad_x, grad_x_csr)


def test_Metrics2d_grad_y_vs_grad_y_csr_rect_diff_res_x_y(
        metric_rect_diff_res_x_y):
    """Check if csr is correctly implemented for grady in diff res rect"""
    metric_rect_diff_res_x_y.check_compute_matrices()
    grad_y = metric_rect_diff_res_x_y.grad_y_slow.toarray()
    grad_y_csr = metric_rect_diff_res_x_y.grad_y_csr.toarray()
    assert np.allclose(grad_y, grad_y_csr)


def test_Metrics2d_grad_x_vs_grad_x_csr_cyl_diff_res_x_y(
        metric_cyl_diff_res_x_y):
    """Check if csr is correctly implemented for gradx in diff res cyl"""
    metric_cyl_diff_res_x_y.check_compute_matrices()
    grad_x = metric_cyl_diff_res_x_y.grad_x_slow.toarray()
    grad_x_csr = metric_cyl_diff_res_x_y.grad_x_csr.toarray()
    assert np.allclose(grad_x, grad_x_csr)


def test_Metrics2d_grad_y_vs_grad_y_csr_cyl_diff_res_x_y(
        metric_cyl_diff_res_x_y):
    """Check if csr is correctly implemented for grady in diff res cyl"""
    metric_cyl_diff_res_x_y.check_compute_matrices()
    grad_y = metric_cyl_diff_res_x_y.grad_y_slow.toarray()
    grad_y_csr = metric_cyl_diff_res_x_y.grad_y_csr.toarray()
    assert np.allclose(grad_y, grad_y_csr)



"""
lay_and_temp_manager.py

Functions which deal with layouts and templates
"""

from pkg_resources import resource_listdir, resource_string

from yaml import safe_load

from showy.utils import decompact_template  # backward compatibility


__all__ = ["fetch_avbp_layouts", "fetch_avbp_templates", "decompact_template"]


def fetch_avbp_layouts():
    """It returns all the avbp layouts in a dictionary

    Output:
    -------
    avbp_layouts : nested object
    """

    avbp_layouts = {}
    layouts_files_list = resource_listdir('pyavbp.visu', 'avbp_layouts')

    for layout_file_name in layouts_files_list:
        layout_string = resource_string('pyavbp.visu', ("avbp_layouts/"
                                                        + layout_file_name))
        layout = safe_load(layout_string)
        avbp_layouts[layout["title"]] = layout

    return avbp_layouts


def fetch_avbp_templates():
    """It returns all the avbp templates in a dictionary

    Output:
    -------
    avbp_templates : nested object
    """

    avbp_templates = {}
    templates_files_list = resource_listdir('pyavbp.visu', 'avbp_templates')

    for template_file_name in templates_files_list:
        template_string = resource_string('pyavbp.visu', ("avbp_templates/"
                                                          + template_file_name))
        template = safe_load(template_string)
        avbp_templates[template["title"]] = template

    return avbp_templates



"""Extract features from a curve"""

import numpy as np

def reptile(arr, axis, pos):
    """expand  "row" POS from axis AXIS of an array ARR everywhere else

    Args:
        arr (np.array): source array
        axis (int): index of axis to expand
        pos (int): index of element to expand
        
    Returns:
        np.array: Expanded array

    >>> a = np.array([[0, 1, 0, 0, 0],
                      [0, 1, 1, 1, 1],
                      [0, 1, 2, 3, 4]])
    >>> reptile(a, axis=1, pos=3)
    array([[0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1],
           [3, 3, 3, 3, 3]])


    """
    sss = [slice(None) for _ in range(arr.ndim)]
    sss[axis] = np.s_[pos:pos+1:]
    return np.repeat(arr[tuple(sss)], arr.shape[axis], axis=axis)


def amax_keepshape(arr, axis):
    """expand  "row" POS from axis AXIS of an array ARR everywhere else

    Args:
        arr (np.array): source array
        axis (int): index of axis to expand
        
    Returns:
        np.array: amax on the same shape

    >>> a = np.array([[0, 1, 0, 0, 0],
                      [0, 1, 1, 1, 1],
                      [0, 1, 2, 3, 3]])
    >>> amax_keepshape(a, axis=1, pos=3)
    array([[1, 1, 1, 1, 1],
          [1, 1, 1, 1, 1],
          [3, 3, 3, 3, 3]])


    """

    return np.repeat(arr.max(axis=axis, keepdims=True), arr.shape[axis], axis=axis)


def amin_keepshape(arr, axis):
    """expand  "row" POS from axis AXIS of an array ARR everywhere else

    Args:
        arr (np.array): source array
        axis (int): index of axis to expand
        
    Returns:
        np.array: amin on the same shape

    >>> a = np.array([[0, 1, 0, 0, 0],
                     [0, 1, -1, 1, 1],
                     [1, 1, 2, 3, 4]])
    >>> amin_keepshape(a, axis=1, pos=3)
    array([[0, 0, 0, 0, 0],
           [-1, -1, -1, -1, -1],
           [1, 1, 1, 1, 1]])

    """
    return np.repeat(arr.min(axis=axis, keepdims=True), arr.shape[axis], axis=axis)





def _end_match(arr_, kappa=0.2, axis=0, end_index=None):
    """Return a linear approx of the  input curve,
    matching exactly at x=x_ref and y=(1-kappa) x_ref. 

    Args:
        arr_ (np.array): source array
        kappa (float): ]0,1[, position of the first point to place the linear match
        axis (int): index of axis to search
        end_index (int):Index of second point to search the linear match (If None, last index)
        
        
    Returns:
        np.array: linear match of source array

        value                                                   
        ^                                                   
        |                                                   
        |               |\                                  
        |   *          /  \                                 
        |         *   /    \                                
        |             |*    -                               
        |            /      *---                            
        |           /           \*---                       
        |          /             .   \----                  
        |          |             .     -  \----
        ----------/              .             \----
        |                        .                  \-*
        |                        .                    .    *
        |                        .                    .
        |                        .                    .
        |                        .                    .
        |----------------------------------------------> axis
                    int((1-kappa)end_index)            end_index
    """
    size =  arr_.shape[axis]

    if end_index is not None:
        idx2 = end_index
    else:
        idx2 = size-1
    idx1 = min(int((1-kappa)*idx2), idx2-1)
    
    y2 = reptile(arr_,axis, idx2)
    y1 = reptile(arr_,axis, idx1)
    
    slope = (y2-y1)/(idx2-idx1)
    adjust = y2 - slope*idx2
    
    x_vals = np.indices(arr_.shape)[axis]
    return slope*x_vals + adjust


def _mask_deviation(arr_, eps, axis, gain=1.):
    """Create a mask true where a fluctuation array far from zero.

    Args:
        arr_ (np.array): source array
        eps (float): ]0,1[ relative fluctuation thresold (0.1 is 10% of max curve amplitude)
        axis (int): index of axis to search
        gain (float): >0 increase or decrease the with of the layer found 
        
        
        
    Returns:
        np.array: masked array equal to one on the fluctuation side.

        value                                                                                
        ^                                                                                
        |                                                                               
        |  x                                                                            
        | x  x                                                                          
   eps  |...............................................                                 
        |x    x                                                                         
        x       x                                                                      
    0    -----------x. x. x. x. x. x. x. x. x  x. x -----> axis                              
                                                                                       
                                                                                       
    -eps ................................................                                 
                                                                                       
                                                                         
        111111000000000000000000000000000000000000000     

    Notes:
        The fluctuation array must vanish to zero toward the end of AXIS
        The thresold EPS is relative to the local ampliture of the 1D vector.
        The wall node is always one in the outpud

                                                                
    """
    
    x_vals = np.indices(arr_.shape)[axis]
    max_index = arr_.shape[axis]-1
    wall = np.where(x_vals==0, 1, 0)
    
    arr_calib = arr_ + 1.e-12*wall
    amp = amax_keepshape(arr_calib, axis=axis)-amin_keepshape(arr_calib, axis=axis)
    
    mask = np.where(np.abs(arr_calib)>eps*amp, 1, 0)
    wh_ = np.where(mask==1, x_vals, 0)
    thresold = amax_keepshape(wh_, axis=axis)
    thresold = np.rint(thresold*gain).astype(int) # apply the gain
    deviation_mask = np.where(x_vals<=np.clip(thresold,0,max_index), 1, 0)
    return deviation_mask

def mask_boundary_layer(source, eps=0.1, gain=1., axis=0):
    """Identify the cells relative to a boundary layer in a numpy multi-d array.
    If no fluctuation detected, only the "wall" node is set to 1.

    Args:
        source (np.array): source array
        eps (float): ]0,1[ relative fluctuation thresold (0.1 is 10% of max curve amplitude)
        gain (float): >0 increase of decrease the width of the layer found 
        axis (int): index of axis to search
        
        
    Returns:
        np.array: masked array equal to one on the fluctuation side.
    
    >>> a = np.array(
                   [[ 0,  1,  0,  0,  0]
                    [ 0,  1,  1,  1,  1]
                    [ 0,  3,  2,  3,  4]
                    [ 0,  1, -2,  3,  4]
                    [ 0,  0,  0,  0,  0]
                    [ 0,  1,  2,  3,  4]]
        )
    >>> mask_boundary_layer(ource, eps=0.1, axis=1)
    array(
        [[1, 1, 0, 0, 0],
        [1, 0, 0, 0, 0],
        [1, 1, 0, 0, 0],
        [1, 1, 1, 0, 0],
        [1, 0, 0, 0, 0],
        [1, 0, 0, 0, 0]]
    )
    
    """

    endm =  _end_match(source, axis=axis)
    fluctu = source - endm
    mask = _mask_deviation(fluctu, eps=eps, axis=axis, gain=gain)
    return mask


""" module to test de phys parts of Arnica """

import numpy as np

from arnica.phys.solid_material import SolidMaterial
from arnica.phys.thermodyn_properties import h_kader
from arnica.phys.thermodyn_properties import h_stanton
from arnica.phys.thermodyn_properties import stanton_colburn
from arnica.phys.wall_thermal_equilibrium import compute_equilibrium

DIM1 = 5


def test_compute_equilibrium():
    """ simple equilibrium for the ceramic + metal case"""

    cold_h = np.ones(DIM1) * 1000.
    hot_h = 1000.
    hot_t_ad = 2000.
    cold_t_ad = 1000.

    layer1 = SolidMaterial(lambda_poly=[10., 0],
                           lambda_range=[100., 4000.])
    ep_layer1 = 2.e-4
    layer2 = SolidMaterial(lambda_poly=[10., 0.],
                           lambda_range=[100., 4000.])
    ep_layer2 = 2.e-4
    _, t_eq, _ = compute_equilibrium(
        hot_t_ad,
        cold_t_ad,
        hot_h,
        cold_h,
        layer1,
        layer2,
        ep_layer1,
        ep_layer2)

    np.testing.assert_allclose(t_eq, np.ones(DIM1) * 1500.)


def test_h_kader():
    """ test the h as in kader """

    t_wall = np.ones(DIM1) * 1000.
    rho_wall = 1.
    y_wall = 1.e-3
    u_2 = 40.
    t_2 = 1400
    temp_adiab = 1400

    h_wall_expected = 273.12113446
    h_wall = h_kader(
        t_wall,
        rho_wall,
        y_wall,
        u_2,
        t_2,
        temp_adiab)

    np.testing.assert_allclose(h_wall, h_wall_expected)


def test_h_stanton():
    """ Check H from stanton"""

    ones = np.ones(DIM1)
    stanton = ones
    rhou = 2 * ones
    c_p = 3 * ones

    h_wall = h_stanton(stanton, rhou, c_p)
    h_wall_expected = 6. * ones

    np.testing.assert_allclose(h_wall, h_wall_expected)


def test_stanton_colburn():
    """ Check Stanton from colburn"""

    reynolds = 10000. * np.ones(DIM1)
    prandtl = 0.7 * np.ones(DIM1)
    

    st_0 = stanton_colburn(np.zeros_like(reynolds), prandtl)
    st_1 = stanton_colburn(reynolds, prandtl)
    print(st_1)
    np.testing.assert_allclose(st_0, 0.26624088*np.ones(DIM1), rtol=1e-6)
    np.testing.assert_allclose(st_1, 0.00699792*np.ones(DIM1), rtol=1e-6)




import numpy as np
from arnica.utils.axipointcloud import AxiPointCloud


def test_axipointcloud():
    """Testing the axipoint cloud utility"""

    pnt_cloud = AxiPointCloud( 
        xcoor=[0, 1, 2],
        ycoor=[0, 1, 2],
        zcoor=[0, 0, 0],
        vars={
            "var1": [1,1,1],
            "var2": [2,2,2],
        }
    )
    
    str_ = """
AxiPointCloud object "Unnamed"

 Angle section (deg.):0.0

 Coordinates:
-X coor (m)----
shape :(3,)
min/max :0/2
-Y coor (m)----
shape :(3,)
min/max :0/2
-Z coor (m)----
shape :(3,)
min/max :0/0
-R coor (m)----
shape :(3,)
min/max :0.0/2.0
-Theta coor (deg)----
shape :(3,)
min/max :0.0/0.0

 Data:
-var1----
shape :(3,)
min/max :1/1
-var2----
shape :(3,)
min/max :2/2"""
    assert pnt_cloud.__str__() == str_

    assert np.allclose(
        pnt_cloud.xyz(),
        np.array(
            [[0, 0, 0],
            [1, 1, 0],
            [2, 2, 0],]
        )
    )

    assert np.allclose(
        pnt_cloud.vars_stack(),
        np.array(
            [[1, 2],
            [1, 2],
            [1, 2],]
        )
    )
    assert np.allclose(
        pnt_cloud.rad(),
        np.array([0.,1.,2.])
    )
    pnt_cloud.recompute_theta_range_from_coords()
    
    assert pnt_cloud.theta_range == 0.

    pnt_cloud.rotate(np.deg2rad(90.))
    
    assert np.allclose(
        pnt_cloud.xyz(),
        np.array(
            [[0., 0., 0.],
            [1., 0., 1.],
            [2., 0., 2.],]
        )
    )

    pnt_cloud.theta_range = np.deg2rad(90.)
    pnt_cloud.dupli_rotate(1)
    
    
    assert np.allclose(
        pnt_cloud.xyz(),
        np.array(
            [[0., 0., 0.],
            [1., 0., 1.],
            [2., 0., 2.],
            [0., 0., 0.],
            [1., -1., 0.],
            [2., -2., 0.],]
        )
    )
    assert pnt_cloud.theta_range == np.deg2rad(180.)

    print(pnt_cloud.theta_range)
    





# here for backward compatibility, use `showy` instead

import warnings

import showy as showy_pkg


__all__ = ["showy", "display"]


def display(**kwargs):
    """Retro compatibility"""

    warnings.warn("Use *showy* instead of *display*", DeprecationWarning)
    showy(**kwargs)


def showy(layout, data, data_c=None, show=True):
    warnings.warn("Use *showy* package instead", DeprecationWarning)
    dataframes = [data]
    if data_c is not None:
        dataframes.append(data_c)

    return showy_pkg.showy(layout, dataframes, show=show)



""" Module to test the dump of dicts into files """
import os
import pytest
import numpy as np
import h5py
from arnica.utils.datadict2file import (dump_dico_2d_nparrays,
                                        dump_dico_2d_time_nparrays,
                                        dump_dico_1d_nparrays,
                                        dump_dico_0d,
                                        plot_dict_data_as_file)
DIM1 = 5
DIM2 = 11

def test_dump_dico_2d_nparrays(datadir):
    """ test of dumping of dico of 2d nparrays into file"""

    x_arr = np.linspace(0., 1., DIM2)
    y_arr = np.linspace(0., 1., DIM2)
    z_arr = np.linspace(0., 1., DIM2)
    grid_3d = np.stack(np.meshgrid(x_arr, y_arr, z_arr), axis=-1)
    grid_2d = grid_3d[:, :, 0, :]

    case = datadir
    path_actu = os.getcwd()
    os.chdir(case)

    data_dict = {}
    dump_dico_2d_nparrays(data_dict,
                          "tmp_2d.h5",
                          np.take(grid_2d, 0, -1),
                          np.take(grid_2d, 1, -1),
                          np.take(grid_2d, 2, -1))
    size_target = DIM2 ** 2
    with h5py.File("tmp_2d.h5", 'r') as fin:
        assert fin['mesh/x'][()].size == size_target
        assert fin['mesh/y'][()].size == size_target
        assert fin['mesh/z'][()].size == size_target

    data_dict["field_1"] = np.ones((DIM2, DIM2, DIM2))
    data_dict["field_2"] = np.zeros((DIM2, DIM2, DIM2))
    dump_dico_2d_nparrays(data_dict,
                          "tmp_3d.h5",
                          np.take(grid_3d, 0, -1),
                          np.take(grid_3d, 1, -1),
                          np.take(grid_3d, 2, -1))
    size_target = DIM2 ** 3
    with h5py.File("tmp_3d.h5", 'r') as fin:
        assert fin['mesh/x'][()].size == size_target
        assert fin['mesh/y'][()].size == size_target
        assert fin['mesh/z'][()].size == size_target
        assert fin['variables/field_1'][()].size == size_target
        assert fin['variables/field_2'][()].size == size_target

    os.chdir(path_actu)

def test_dump_dico_2d_time_nparrays(datadir):
    """ test of dumping of dico of 2d nparrays along time into files"""

    grid = np.zeros((DIM2, DIM2, 3))

    data_dict = {}
    data_dict["field"] = np.ones((DIM1, DIM2, DIM2)) 

    case = datadir
    path_actu = os.getcwd()
    os.chdir(case)

    dump_dico_2d_time_nparrays(data_dict,
                               "./",
                               "tmp",
                               np.take(grid, 0, -1),
                               np.take(grid, 1, -1),
                               np.take(grid, 2, -1))

    for step in range(DIM1):
        assert os.path.isfile(f"./tmp_{step:08}.h5")
        size_target = DIM2 ** 2
        with h5py.File(f"./tmp_{step:08}.h5", 'r') as fin:
            assert fin['mesh/x'][()].size == size_target
            assert fin['mesh/y'][()].size == size_target
            assert fin['mesh/z'][()].size == size_target
            assert fin['variables/field'][()].size == size_target
    assert os.path.isfile(f"./tmp_collection.xmf")

    os.chdir(path_actu)

def test_dump_dico_1d_nparrays(datadir):
    """ test of dumping of dico of 1d nparrays into xlsx or csv file"""

    data_dict = {}
    data_dict["field_1"] = np.ones(DIM2)
    data_dict["field_2"] = np.ones(DIM2)

    case = datadir
    path_actu = os.getcwd()
    os.chdir(case)

    filename = "./dir1/tmp_1d.csv"
    dump_dico_1d_nparrays(filename, data_dict)
    assert os.path.isfile(filename)
    with open(filename, 'r') as fin:
        lines = fin.readlines()
    assert lines[0] == "# field_1 , field_2\n"
    assert len(lines) == DIM2 + 1

    filename = "./dir1/tmp_1d.xlsx"
    try:
        import pandas
        import openpyxl
        packages_exist = True
    except ImportError:
        packages_exist = False
    if packages_exist:
        dump_dico_1d_nparrays(filename, data_dict)
        assert os.path.isfile(filename)
    else:
        with pytest.raises(ImportError):
            dump_dico_1d_nparrays(filename, data_dict)

    os.chdir(path_actu)

def test_dump_dico_0d(datadir):
    """ test of dumping of dico of 0d float into xlsx or csv file"""

    data_dict = {}
    data_dict['float_1'] = 4.
    data_dict['float_2'] = 2.

    case = datadir
    path_actu = os.getcwd()
    os.chdir(case)

    filename = "./dir2/tmp_0d.csv"
    dump_dico_0d(filename, data_dict)
    assert os.path.isfile(filename)
    with open(filename, 'r') as fin:
        lines = fin.readlines()
    assert lines[0] == "# float_1 , float_2\n"
    assert len(lines) == 2

    filename = "./dir2/tmp_0d.xlsx"
    try:
        import pandas
        import openpyxl
        packages_exist = True
    except ImportError:
        packages_exist = False
    if packages_exist:
        dump_dico_0d(filename, data_dict)
        assert os.path.isfile(filename)
    else:
        with pytest.raises(ImportError):
            dump_dico_0d(filename, data_dict)

    os.chdir(path_actu)

def test_plot_dict_data_as_file(datadir):
    """ test of the plot of dict data """

    filename = "./dir3/tmp_plot.dummy_format"
    data_dict = {}
    data_dict["x_axis"] = np.linspace(0, DIM2 - 1, DIM2)
    data_dict["y_axis"] = np.ones(DIM2)

    case = datadir
    path_actu = os.getcwd()
    os.chdir(case)

    plot_dict_data_as_file(data_dict, filename, 'x_axis', 'y_axis')
    assert os.path.isfile("./dir3/tmp_plot.pdf")

    os.chdir(path_actu)




# here for backward compatibility, use `h5cross` instead

from h5cross.same_nob import h5_same
from h5cross.same_nob import dict_same
from h5cross.same_nob import h5dict_safe


__all__ = ["h5_same", "dict_same", "h5dict_safe"]





import numpy as np
from arnica.utils.show_mat import show_mat


def test_show_mat():
    def samplemat(dims):
        """Make a matrix with all zeros and increasing elements 
        on the diagonal ans last dim"""
        aa = np.zeros(dims)
        for i in range(min(dims)):
            aa[i, :i] = i
        return aa
    # Display matrix
    show_mat(samplemat((15, 15)), "trout", show=False)





import pytest
from arnica.phys.yk_from_phi import yk_from_phi


def test_ykfromphi():
    yf_methane = yk_from_phi(1.0, 1, 4)["fuel"]
    yf_c3h8 = yk_from_phi(1.0, 3, 8)["fuel"]
    yf_c8h18 = yk_from_phi(1.0, 3, 8)["fuel"]
    yf_h2 = yk_from_phi(1.0, 3, 8)["fuel"]
    pytest.approx(yf_methane, 0.001) == 0.055
    pytest.approx(yf_c3h8, 0.001) == 0.06
    pytest.approx(yf_c8h18, 0.001) == 0.062
    pytest.approx(yf_h2, 0.001) == 0.028
